/* 
   divx4file_core.c (C) Justin Schoeman 2001 (justin@suntiger.ee.up.ac.za)

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
    
*/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <stdint.h>
#include <dlfcn.h>
#include "avilib.h"
#include "mcs.h"
#include "encore2.h"
#include "lame3.70/lame.h"
#include "../NVrec.h"

#define MP3BUFSIZE 65536

struct divx4file_core_t {
	avi_t * avi;
	mcs_t * mcs;
	int (*encore)(void *handle, int enc_opt, void *param1, void *param2);
	int divx_version;
	void * dh;
	void * dlh;
	lame_global_flags lgf;
	unsigned char * cbuf;
	unsigned char mbuf[MP3BUFSIZE];
	ENC_FRAME enc_frame;
	ENC_RESULT enc_result;
	int ashift;
	int aformat;
	int vformat;
	uint64_t atime;
	uint64_t vtime;
	uint64_t tstep;
	int drop;
};

struct file_core_t divx4file_core;

void divx4file_core_audio(struct file_core_t * fct, unsigned char *buf, int n)
{
	struct divx4file_core_t * divx4c = (struct divx4file_core_t *)fct->priv;
	int i = 0;
	
	if(!divx4c)
		return;
	if(divx4c->avi) {
		if(divx4c->aformat == 0x55) {
			i = lame_encode_buffer_interleaved(&divx4c->lgf, (short int *)buf, n >> divx4c->ashift, divx4c->mbuf, MP3BUFSIZE);
			if(i > 0)
				AVI_write_audio(divx4c->avi, divx4c->mbuf, i);
		} else if(divx4c->aformat == WAVE_FORMAT_PCM) {
			AVI_write_audio(divx4c->avi, buf, n);
		}
	} else {
		if(divx4c->aformat == 0x55) {
			i = lame_encode_buffer_interleaved(&divx4c->lgf, (short int *)buf, n >> divx4c->ashift, divx4c->mbuf, MP3BUFSIZE);
			if(i > 0)
				MCS_write_audio(divx4c->mcs, divx4c->mbuf, i, divx4c->atime);
		} else if(divx4c->aformat == WAVE_FORMAT_PCM) {
			MCS_write_audio(divx4c->mcs, buf, n, divx4c->atime);
		}
	}
	divx4c->atime += divx4c->tstep;
}

void divx4file_core_close(struct file_core_t * fct)
{
	struct divx4file_core_t * divx4c = (struct divx4file_core_t *)fct->priv;
	int i;
	
	if(!divx4c)
		return;
	if(divx4c->encore(divx4c->dh, ENC_OPT_RELEASE, NULL, NULL) != ENC_OK)
		fprintf(stderr, "Error freeing encore device!\n");
	if(divx4c->aformat == 0x55) {
		i = lame_encode_finish(&divx4c->lgf, divx4c->mbuf, MP3BUFSIZE);
		if(i > 0) {
			if(divx4c->avi)
				AVI_write_audio(divx4c->avi, divx4c->mbuf, i);
			else
				MCS_write_audio(divx4c->mcs, divx4c->mbuf, i, divx4c->atime);
		}
	}
	if(divx4c->avi)
		AVI_close(divx4c->avi);
	if(divx4c->mcs)
		MCS_close(divx4c->mcs);
	dlclose(divx4c->dlh);
	free(divx4c);
	free(fct);
}

void divx4file_core_nullvideo(struct file_core_t * fct)
{
	struct divx4file_core_t * divx4c = (struct divx4file_core_t *)fct->priv;
	
	if(!divx4c)
		return;
	if(divx4c->avi)
		AVI_write_frame(divx4c->avi, divx4c->cbuf, 0, AVI_FLAG_NONE);
	else
		MCS_write_frame(divx4c->mcs, divx4c->cbuf, 0, divx4c->vtime);
	divx4c->vtime += divx4c->tstep;
}

void divx4file_core_video(struct file_core_t * fct, unsigned char **buf)
{
	struct divx4file_core_t * divx4c = (struct divx4file_core_t *)fct->priv;
	
	if(!divx4c)
		return;
	memset(&divx4c->enc_frame, 0, sizeof(ENC_FRAME));
	divx4c->enc_frame.image = buf[0];
	divx4c->enc_frame.bitstream = divx4c->cbuf;
	divx4c->enc_frame.colorspace = divx4c->vformat;
	if(divx4c->encore(divx4c->dh, ENC_OPT_ENCODE, &divx4c->enc_frame, &divx4c->enc_result) != ENC_OK) {
		fprintf(stderr, "Error encoding frame!\n");
		if(divx4c->avi)
			AVI_write_frame(divx4c->avi, divx4c->cbuf, 0, AVI_FLAG_NONE);
		else
			MCS_write_frame(divx4c->mcs, divx4c->cbuf, 0, divx4c->vtime);
	} else {
		if(divx4c->avi)
			AVI_write_frame(divx4c->avi, 
					divx4c->cbuf, 
					divx4c->enc_frame.length, 
					divx4c->enc_result.is_key_frame ? 
						AVI_FLAG_KEYFRAME : AVI_FLAG_NONE);
		else
			MCS_write_frame(divx4c->mcs, 
					divx4c->cbuf, 
					divx4c->enc_frame.length, 
					divx4c->vtime);
	}
	divx4c->vtime += divx4c->tstep;
}

void divx4file_core_sync(struct file_core_t * fct)
{
	return;
}

static int divx4file_core_formats[] = {NV_YUV420, NV_YUV422, NV_RGB24, NV_END_FORMAT};

unsigned int *divx4file_core_format(char **args)
{
	return divx4file_core_formats;
}

void divx4file_core_usage()
{
	fprintf(stderr, "DIVX4 file core options:\n");
	fprintf(stderr, "\t-ab <audio mp3 bitrate (kbps)>               [?]\n");
	fprintf(stderr, "\t-aq <audio mp3 quality factor: 0..100>       [?]\n");
	fprintf(stderr, "\t-wav use raw PCM instead of mp3              [mp3]\n");
	fprintf(stderr, "\t-xvid <use xvid codec>                       [DivX;)]\n");
	fprintf(stderr, "\t-xvid_divx <use xvid codec with DivX fourcc> [DivX;)]\n");
	fprintf(stderr, "\t-vb <divx bitrate (kbps)>                    [?]\n");
	fprintf(stderr, "\t-vq <divx quality factor: 0..100>            [?]\n");
	fprintf(stderr, "\t-vg <maximum key interval>                   [?]\n");
	fprintf(stderr, "\t-dqmin <DivX minimum quantiser>              [?]\n");
	fprintf(stderr, "\t-dqmax <DivX maximum quantizer>              [?]\n");
	fprintf(stderr, "\t-drcp <DivX rate control period>             [?2000]\n");
	fprintf(stderr, "\t-drcrp <DivX rate control reaction period>   [?10]\n");
	fprintf(stderr, "\t-drcrr <DivX rate control reaction ratio>    [?20]\n");
	fprintf(stderr, "\t-deinterlace <deinterlace> *BROKEN*          [off]\n");
#if (ENCORE_MAJOR_VERSION > 5000)
	fprintf(stderr, "\t-dbidir enable DivX B frames                 [off]\n");
	fprintf(stderr, "\t-ddint enable DivX deinterlacer              [off]\n");
	fprintf(stderr, "\t-dtfilt <DivX temporal filter level: 0.0-1.0>[off]\n");
	fprintf(stderr, "\t-dsfilt <DivX spatial filter level: 0.0-1.0> [off]\n");
	fprintf(stderr, "\t-dspass <DivX spatial filter passes: 1-3>    [off]\n");
#endif
	fprintf(stderr, "\t-o <output filename>                         [NONE]\n");
	fprintf(stderr, "\t * filename starting with udp:// indicates multicast.\n");
	fprintf(stderr, "\t * filename starting with tcp:// indicates unicast.\n");
}

struct file_core_t * divx4file_core_init(char ** args, unsigned int format, ntime_t vrate)
{
	struct file_core_t * fct;
	struct divx4file_core_t * divx4c;
	double f;
 	int arate = -1;
 	int abits = -1;
 	int achans = -1;
 	int abr = -1;
 	int aq = -1;
 	int vwidth = -1;
 	int vheight = -1;
 	char enclib[50];
 	char fcc[5];
 	char *fname = NULL;
 	ENC_PARAM enc_param;
 	NV_ARGS;
 	
 	fct =  malloc(sizeof(struct file_core_t));
 	if(!fct) {
 		perror("divx4file core init - could no malloc new cap core");
 		return NULL;
 	}
 	divx4c = malloc(sizeof(struct divx4file_core_t));
 	if(!divx4c) {
 		perror("divx4file core init - could no malloc new divx4 core");
 		free(fct);
 		return NULL;
 	}
 	memcpy(fct, &divx4file_core, sizeof(struct file_core_t));
 	fct->priv = (void *)divx4c;
 	memset(divx4c, 0, sizeof(struct divx4file_core_t));
 	divx4c->aformat = 0x55;

 	strncpy(enclib, "libdivxencore.so", 50);
 	strncpy(fcc, "DIVX", 5);
 	
 	switch(format) {
 		case NV_RGB24:	divx4c->vformat = ENC_CSP_RGB24; break;
 		case NV_YUV420:	divx4c->vformat = ENC_CSP_I420; break;
 		case NV_YUV422:	divx4c->vformat = ENC_CSP_YUY2; break;
 		default:
 			fprintf(stderr, "divx4file core init - invalid colour format!\n");
 			goto error0;
 	}
 	
	memset(&enc_param, 0, sizeof(ENC_PARAM));
	NV_SCAN_ARGS {
		if(strcmp(NV_THIS_ARG, "-r") == 0) {
			if(!NV_NEXT_ARG)
				goto usage;
			arate = atoi(NV_THIS_ARG);
		} else if(strcmp(NV_THIS_ARG, "-b") == 0) {
			if(!NV_NEXT_ARG)
				goto usage;
			abits = atoi(NV_THIS_ARG);
		} else if(strcmp(NV_THIS_ARG, "-s") == 0) {
			achans = 2;
		} else if(strcmp(NV_THIS_ARG, "-m") == 0) {
			achans = 1;
		} else if(strcmp(NV_THIS_ARG, "-w") == 0) {
			if(!NV_NEXT_ARG)
				goto usage;
			vwidth = atoi(NV_THIS_ARG);
		} else if(strcmp(NV_THIS_ARG, "-h") == 0) {
			if(!NV_NEXT_ARG)
				goto usage;
			vheight = atoi(NV_THIS_ARG);
		} else if(strcmp(NV_THIS_ARG, "-ab") == 0) {
			if(!NV_NEXT_ARG)
				goto usage;
			abr = atoi(NV_THIS_ARG);
			if(abr <= 0)
				goto usage;
		} else if(strcmp(NV_THIS_ARG, "-aq") == 0) {
			if(!NV_NEXT_ARG)
				goto usage;
			aq = atoi(NV_THIS_ARG);
			aq /= 10;
			aq = (aq > 9) ? 9 : aq;
			aq = 9 - aq;
			if((aq < 0) || (aq > 9))
				goto usage;
		} else if(strcmp(NV_THIS_ARG, "-vb") == 0) {
			if(!NV_NEXT_ARG)
				goto usage;
			enc_param.bitrate = atoi(NV_THIS_ARG);
			if(enc_param.bitrate <= 0)
				goto usage;
			enc_param.bitrate *= 1000;
		} else if(strcmp(NV_THIS_ARG, "-vq") == 0) {
			if(!NV_NEXT_ARG)
				goto usage;
			enc_param.quality = atoi(NV_THIS_ARG);
			enc_param.quality += 20;
			enc_param.quality /= 20;
			enc_param.quality =
				(enc_param.quality > 5) ? 5 : enc_param.quality;
			if((enc_param.quality < 1) || (enc_param.quality > 5))
				goto usage;
		} else if(strcmp(NV_THIS_ARG, "-dqmin") == 0) {
			if(!NV_NEXT_ARG)
				goto usage;
			enc_param.min_quantizer = atoi(NV_THIS_ARG);
			if(enc_param.min_quantizer <= 0)
				goto usage;
		} else if(strcmp(NV_THIS_ARG, "-dqmax") == 0) {
			if(!NV_NEXT_ARG)
				goto usage;
			enc_param.max_quantizer = atoi(NV_THIS_ARG);
			if(enc_param.max_quantizer <= 0)
				goto usage;
		} else if(strcmp(NV_THIS_ARG, "-vg") == 0) {
			if(!NV_NEXT_ARG)
				goto usage;
			enc_param.max_key_interval = atoi(NV_THIS_ARG);
			if(enc_param.max_key_interval <= 0)
				goto usage;
		} else if(strcmp(NV_THIS_ARG, "-drcp") == 0) {
			if(!NV_NEXT_ARG)
				goto usage;
			enc_param.rc_period = atoi(NV_THIS_ARG);
			if(enc_param.rc_period <= 0)
				goto usage;
		} else if(strcmp(NV_THIS_ARG, "-drcrp") == 0) {
			if(!NV_NEXT_ARG)
				goto usage;
			enc_param.rc_reaction_period = atoi(NV_THIS_ARG);
			if(enc_param.rc_reaction_period <= 0)
				goto usage;
		} else if(strcmp(NV_THIS_ARG, "-drcrr") == 0) {
			if(!NV_NEXT_ARG)
				goto usage;
			enc_param.rc_reaction_ratio = atoi(NV_THIS_ARG);
			if(enc_param.rc_reaction_ratio <= 0)
				goto usage;
		} else if(strcmp(NV_THIS_ARG, "-deinterlace") == 0) {
			enc_param.deinterlace = 1;
		} else if(strcmp(NV_THIS_ARG, "-wav") == 0) {
			divx4c->aformat = WAVE_FORMAT_PCM;
		} else if(strcmp(NV_THIS_ARG, "-o") == 0) {
			if(!NV_NEXT_ARG)
				goto usage;
			fname = NV_THIS_ARG;
		} else if(strcmp(NV_THIS_ARG, "-xvid") == 0) {
		 	strncpy(enclib, "libxvidcore.so", 50);
		 	strncpy(fcc, "XVID", 5);
		} else if(strcmp(NV_THIS_ARG, "-xvid_divx") == 0) {
		 	strncpy(enclib, "libxvidcore.so", 50);
		 	strncpy(fcc, "DIVX", 5);
		} else if(strcmp(NV_THIS_ARG, "-vr") == 0) {
			if(!NV_NEXT_ARG)
				goto usage;
			vrate = (unsigned long long)(1e9 / atof(NV_THIS_ARG));
		}
#if (ENCORE_MAJOR_VERSION > 5000)
		else if(divx4c->divx_version > 5000) {
			if(strcmp(NV_THIS_ARG, "-dbidir") == 0) {
				enc_param.extensions.use_bidirect = 1;
			} else if(strcmp(NV_THIS_ARG, "-ddint") == 0) {
				enc_param.extensions.interlace_mode = 1;
			} else if(strcmp(NV_THIS_ARG, "-dtfilt") == 0) {
				enc_param.extensions.temporal_enable = 1;
				if(!NV_NEXT_ARG)  
					goto usage;
				enc_param.extensions.temporal_level = atof(NV_THIS_ARG);
			} else if(strcmp(NV_THIS_ARG, "-dsfilt") == 0) {
				if(!NV_NEXT_ARG)  
					goto usage;
				enc_param.extensions.spatial_level = atof(NV_THIS_ARG);
			} else if(strcmp(NV_THIS_ARG, "-dspass") == 0) {
				if(!NV_NEXT_ARG)  
					goto usage;
				enc_param.extensions.spatial_passes = atoi(NV_THIS_ARG);
			}
		}
#endif
	}
	divx4c->ashift = (achans == 1) ? 1 : 2;

	if(!fname)
		goto usage;
	fprintf(stderr, "divx4file core init - file: %s (%dx%d)\n", fname, vwidth, vheight);
	
	divx4c->dlh = dlopen(enclib, RTLD_NOW);
	if(!divx4c->dlh) {
		fprintf(stderr, "divx4file core init - loading enore %s\n", dlerror());
		goto error0;
	}

	divx4c->encore = dlsym(divx4c->dlh, "encore");
	if(!divx4c->encore) {
		fprintf(stderr, "divx4file core init - find encore %s\n", dlerror());
		goto error1;
	}
	
#ifdef ENC_OPT_VERSION
	if(strcmp(enclib, "libxvidcore.so") != 0) {
		divx4c->divx_version = divx4c->encore(NULL, ENC_OPT_VERSION, NULL, NULL);
		if(divx4c->divx_version != ENCORE_VERSION) {
			fprintf(stderr, "divx4file core init - encore binary does not match header this was compiled for!\n");
			goto error1;
		}
		divx4c->divx_version = ENCORE_MAJOR_VERSION;
		if(divx4c->divx_version == 5000) {
			fprintf(stderr, "divx4file core init - this divx4 codec does not contain an encoder, please download the latest version.\n");
			goto error1;
		}
		if(divx4c->divx_version > 5000) {
			strncpy(fcc, "DX50", 5);
		}
	} else {
		divx4c->divx_version = 4000;
	}
#else
	divx4c->divx_version = divx4c->encore(NULL, 4, NULL, NULL);
	if(divx4c->divx_version > 20020000) {   
		fprintf(stderr, "divx4file core init - You seem to have a newer divx codec installed - please recompile NVrec\n");
		goto error1;
	}
	divx4c->divx_version = 4000;
#endif
	
	if( (strncmp(fname + 3, "://", 3) == 0) ||
		(strcmp(fname, "-") == 0) ) {
		divx4c->mcs = MCS_open(fname);
		divx4c->avi = NULL;
		if(!divx4c->mcs) {
			perror("divx4file core init - could not open multicast");
			goto error1;
		}
	} else {
		divx4c->avi = AVI_open_output_file(fname);
		divx4c->mcs = NULL;
		if(!divx4c->avi) {
			perror("divx4file core init - could not open file");
			goto error1;
		}
	}
	f = 1e9 / (double)vrate;
	fprintf(stderr, "file core init - frame rate is %f\n", f);

	enc_param.x_dim = vwidth;
	enc_param.y_dim = vheight;
	enc_param.framerate = f;
	if(divx4c->encore(NULL, ENC_OPT_INIT, &enc_param, NULL) != ENC_OK) {
		fprintf(stderr, "file core init - error initing encoder\n");
		goto error1;
	}
	divx4c->dh = enc_param.handle;
	if(!divx4c->dh)
		goto error2;
	
	fprintf(stderr, "Got DivX bitrate: %d\n", enc_param.bitrate);
	
	divx4c->cbuf = malloc(vwidth * vheight * 6);
	if(!divx4c->cbuf) {
		perror("file core init - could not allocate compression buffer");
		goto error3;
	}
	
	if(divx4c->avi)
		AVI_set_video(divx4c->avi, vwidth, vheight, f, fcc);
	else
		MCS_set_video(divx4c->mcs, vwidth, vheight, f, fcc);

	if(divx4c->aformat == 0x55) {
		lame_init(&divx4c->lgf);
		if(abr > 0)
			divx4c->lgf.brate = abr;
		if(aq > 0)
			divx4c->lgf.quality = aq;
		divx4c->lgf.num_channels = achans;
		divx4c->lgf.in_samplerate = arate;
		lame_init_params(&divx4c->lgf);
		if(divx4c->avi)
			AVI_set_audio(divx4c->avi, achans, arate, abits, divx4c->aformat, (divx4c->lgf.brate * 1000) / 8, divx4c->lgf.framesize);
		else
			MCS_set_audio(divx4c->mcs, achans, arate, abits, f, divx4c->aformat);
	} else if(divx4c->aformat == WAVE_FORMAT_PCM) {
		if(divx4c->avi)
			AVI_set_audio(divx4c->avi, achans, arate, abits, divx4c->aformat, 0, 0);
		else
			MCS_set_audio(divx4c->mcs, achans, arate, abits, f, divx4c->aformat);
	} else {
		fprintf(stderr, "Oops - bad audio format!\n");
		goto error3;
	}
	divx4c->tstep = vrate;
	return (void *)fct;

error3:	
	divx4c->encore(divx4c->dh, ENC_OPT_RELEASE, NULL, NULL);
error2:
	if(divx4c->avi)
		AVI_close(divx4c->avi);
	if(divx4c->mcs)
		MCS_close(divx4c->mcs);
error1:
	dlclose(divx4c->dlh);
error0:
	free(fct);
	free(divx4c);
	return NULL;

usage:
	fprintf(stderr, "divx4file core usage:\n");
	fprintf(stderr, "Error parsing options.  Please run without any command line options for\n");
	fprintf(stderr, "detailed usage information.\n");
	free(fct);
	free(divx4c);
	return NULL;
}

void divx4file_core_overrides(char **args)
{
	nvcore_add_arg(args, "-b");
	nvcore_add_arg(args, "16");
	return;
}

struct file_core_t divx4file_core = {
	file_core_init: divx4file_core_init,
	file_core_close: divx4file_core_close,
	file_core_nullvideo: divx4file_core_nullvideo,
	file_core_fastvideo: divx4file_core_video,
	file_core_video: divx4file_core_video,
	file_core_audio: divx4file_core_audio,
	file_core_sync: divx4file_core_sync,
	file_core_format: divx4file_core_format,
	file_core_overrides: divx4file_core_overrides,
	file_core_usage: divx4file_core_usage,
	priv: NULL
};
